/**
* Observer pattern example.
*/

package unleashed.ch21.observer;


import java.util.*;

/**
* Overdraft event interface
*/
interface OverdraftEvent
{
	public double getOverdraftAmount();
}

/**
* Overdraft event listener
*/
interface OverdraftEventListener
{
	public void overdraftHappened(OverdraftEvent event);
	
}

/**
* Overdraft notifier
*/
interface OverdraftNotifier
{
	public void addOverdraftListener(OverdraftEventListener listener);
	public void removeOverdraftListener(OverdraftEventListener listener);
}

/**
* Overdraft event implementation.
*/
class OverdraftEventImpl implements OverdraftEvent
{
	protected double amount;

	public OverdraftEventImpl(double amount)
	{
		this.amount = amount;
	}
	
	public double getOverdraftAmount()
	{
		return amount;
	}
}

/**
* Overdraft listener implementaton
*/
class OverdraftProcessor implements OverdraftEventListener
{
	public void overdraftHappened(OverdraftEvent event)
	{
		//do something about overdraft
		System.out.println("Overdraft event happened " + event.getOverdraftAmount());
	}
	
}

/**
* Account notifies of overdraft
*/
public class Account implements OverdraftNotifier
{

	//account balance 
	protected double balance = 0;

	//create a vector of listeners
	protected Vector listeners = new Vector();

	public Account(double amount)
	{
		balance = amount;
	}
	
	public void addOverdraftListener(OverdraftEventListener listener)
	{
		listeners.add(listener);
	}

	public void removeOverdraftListener(OverdraftEventListener listener)
	{
		listeners.remove(listener);		
	}

	public void withdraw(double amount)
	{
		double newbalance = balance - amount;

		if(balance < amount)
		{
			notifyOverdraft(amount - balance);
		}
		
		balance = balance - amount;
	}
	
	public void notifyOverdraft(double overdraftAmount)
	{
		//create event
		OverdraftEventImpl event = new OverdraftEventImpl(overdraftAmount);

		//notify every listener
		for(Iterator it = listeners.iterator(); it.hasNext();)
		{
		
			OverdraftEventListener listener =  (OverdraftEventListener)it.next();
			System.out.println("Notifying of overdraft");
			listener.overdraftHappened(event);
		}

	}
	
	public static void main(String[] args)
	{
		//create account with balance of 1000
		System.out.println("Creating account with balance of 1000");
		Account account = new Account(1000);
		
		//create overdraft listener
		System.out.println("Creating overdraft listener");
		OverdraftEventListener listener = new OverdraftProcessor();
		
		//add listener to account
		System.out.println("Adding listener to notifier");
		account.addOverdraftListener(listener);
		
		//withdraw some money
		System.out.println("Withdrawing 800");
		account.withdraw(800);
		
		//withdraw more money
		System.out.println("Withdrawing 300");
		account.withdraw(300);
		
		
	}
}
